@@ -1,5 +1,5 @@
 /*  Lunzip - Decompressor for the lzip format
-    Copyright (C) 2010-2016 Antonio Diaz Diaz.
+    Copyright (C) 2010-2017 Antonio Diaz Diaz.
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -53,7 +53,7 @@ void Pp_show_msg( struct Pretty_print * const pp, const char * const msg )
 /* Returns the number of bytes really read.
    If (returned value < size) and (errno == 0), means EOF was reached.
 */
-static int readblock( const int fd, uint8_t * const buf, const int size )
+int readblock( const int fd, uint8_t * const buf, const int size )
   {
   int sz = 0;
   errno = 0;
@@ -87,10 +87,10 @@ static int writeblock( const int fd, const uint8_t * const buf, const int size )
   }
 
 
-int seek_read( const int fd, uint8_t * const buf, const int size,
-               const int offset )
+unsigned seek_read_back( const int fd, uint8_t * const buf, const int size,
+                         const int offset )
   {
-  if( lseek( fd, offset, SEEK_END ) >= 0 )
+  if( lseek( fd, -offset, SEEK_END ) >= 0 )
     return readblock( fd, buf, size );
   return 0;
   }
@@ -121,8 +121,9 @@ void LZd_flush_data( struct LZ_decoder * const d )
         writeblock( d->outfd, d->buffer + d->stream_pos, size ) != size )
       { show_error( "Write error", errno, false ); cleanup_and_fail( 1 ); }
     if( d->pos >= d->buffer_size )
-      { d->partial_data_pos += d->pos; d->pos = 0;
-        if( d->partial_data_pos >= d->dictionary_size ) d->pos_wrapped = true; }
+      { d->partial_data_pos += d->pos; d->pos = 0; d->pos_wrapped = true;
+        if( d->partial_data_pos >= d->dictionary_size )
+          d->pos_wrapped_dic = true; }
     d->stream_pos = d->pos;
     }
   }
@@ -185,7 +186,7 @@ static bool LZd_verify_trailer( struct LZ_decoder * const d,
              ( 8.0 * member_size ) / data_size,
              100.0 * ( 1.0 - ( (double)member_size / data_size ) ) );
   if( !error && verbosity >= 4 )
-    fprintf( stderr, "data CRC %08X, data size %9llu, member size %8llu.  ",
+    fprintf( stderr, "CRC %08X, decompressed %9llu, compressed %8llu.  ",
              LZd_crc( d ), data_size, member_size );
   return !error;
   }
@@ -198,49 +199,77 @@ int LZd_decode_member( struct LZ_decoder * const d,
   {
   struct Range_decoder * const rdec = d->rdec;
   void (* const copy_block)
-       ( struct LZ_decoder * const d, const int distance, int len ) =
-    ( (unsigned)d->buffer_size >= d->dictionary_size ) ?
+       ( struct LZ_decoder * const d, const unsigned distance, unsigned len ) =
+    ( d->buffer_size >= d->dictionary_size ) ?
     &LZd_copy_block : &LZd_copy_block2;
+  Bit_model bm_literal[1<<literal_context_bits][0x300];
+  Bit_model bm_match[states][pos_states];
+  Bit_model bm_rep[states];
+  Bit_model bm_rep0[states];
+  Bit_model bm_rep1[states];
+  Bit_model bm_rep2[states];
+  Bit_model bm_len[states][pos_states];
+  Bit_model bm_dis_slot[len_states][1<<dis_slot_bits];
+  Bit_model bm_dis[modeled_distances-end_dis_model+1];
+  Bit_model bm_align[dis_align_size];
+  struct Len_model match_len_model;
+  struct Len_model rep_len_model;
   unsigned rep0 = 0;		/* rep[0-3] latest four distances */
   unsigned rep1 = 0;		/* used for efficient coding of */
   unsigned rep2 = 0;		/* repeated distances */
   unsigned rep3 = 0;
   State state = 0;
 
+  Bm_array_init( bm_literal[0], (1 << literal_context_bits) * 0x300 );
+  Bm_array_init( bm_match[0], states * pos_states );
+  Bm_array_init( bm_rep, states );
+  Bm_array_init( bm_rep0, states );
+  Bm_array_init( bm_rep1, states );
+  Bm_array_init( bm_rep2, states );
+  Bm_array_init( bm_len[0], states * pos_states );
+  Bm_array_init( bm_dis_slot[0], len_states * (1 << dis_slot_bits) );
+  Bm_array_init( bm_dis, modeled_distances - end_dis_model + 1 );
+  Bm_array_init( bm_align, dis_align_size );
+  Lm_init( &match_len_model );
+  Lm_init( &rep_len_model );
+
   Rd_load( rdec );
   while( !Rd_finished( rdec ) )
     {
     const int pos_state = LZd_data_position( d ) & pos_state_mask;
-    if( Rd_decode_bit( rdec, &d->bm_match[state][pos_state] ) == 0 )	/* 1st bit */
+    if( Rd_decode_bit( rdec, &bm_match[state][pos_state] ) == 0 )	/* 1st bit */
       {
-      const uint8_t prev_byte = LZd_peek_prev( d );
+      Bit_model * const bm = bm_literal[get_lit_state(LZd_peek_prev( d ))];
       if( St_is_char( state ) )
         {
         state -= ( state < 4 ) ? state : 3;
-        LZd_put_byte( d, Rd_decode_tree( rdec,
-                         d->bm_literal[get_lit_state(prev_byte)], 8 ) );
+        LZd_put_byte( d, Rd_decode_tree8( rdec, bm ) );
         }
       else
         {
         state -= ( state < 10 ) ? 3 : 6;
-        LZd_put_byte( d, Rd_decode_matched( rdec,
-                         d->bm_literal[get_lit_state(prev_byte)],
-                         LZd_peek( d, rep0 ) ) );
+        LZd_put_byte( d, Rd_decode_matched( rdec, bm, LZd_peek( d, rep0 ) ) );
         }
       }
     else					/* match or repeated match */
       {
       int len;
-      if( Rd_decode_bit( rdec, &d->bm_rep[state] ) != 0 )	/* 2nd bit */
+      if( Rd_decode_bit( rdec, &bm_rep[state] ) != 0 )		/* 2nd bit */
         {
-        if( Rd_decode_bit( rdec, &d->bm_rep0[state] ) != 0 )	/* 3rd bit */
+        if( Rd_decode_bit( rdec, &bm_rep0[state] ) == 0 )	/* 3rd bit */
+          {
+          if( Rd_decode_bit( rdec, &bm_len[state][pos_state] ) == 0 ) /* 4th bit */
+            { state = St_set_short_rep( state );
+              LZd_put_byte( d, LZd_peek( d, rep0 ) ); continue; }
+          }
+        else
           {
           unsigned distance;
-          if( Rd_decode_bit( rdec, &d->bm_rep1[state] ) == 0 )	/* 4th bit */
+          if( Rd_decode_bit( rdec, &bm_rep1[state] ) == 0 )	/* 4th bit */
             distance = rep1;
           else
             {
-            if( Rd_decode_bit( rdec, &d->bm_rep2[state] ) == 0 )	/* 5th bit */
+            if( Rd_decode_bit( rdec, &bm_rep2[state] ) == 0 )	/* 5th bit */
               distance = rep2;
             else
               { distance = rep3; rep3 = rep2; }
@@ -249,36 +278,29 @@ int LZd_decode_member( struct LZ_decoder * const d,
           rep1 = rep0;
           rep0 = distance;
           }
-        else
-          {
-          if( Rd_decode_bit( rdec, &d->bm_len[state][pos_state] ) == 0 )	/* 4th bit */
-            { state = St_set_short_rep( state );
-              LZd_put_byte( d, LZd_peek( d, rep0 ) ); continue; }
-          }
         state = St_set_rep( state );
-        len = min_match_len + Rd_decode_len( rdec, &d->rep_len_model, pos_state );
+        len = min_match_len + Rd_decode_len( rdec, &rep_len_model, pos_state );
         }
       else					/* match */
         {
-        int dis_slot;
-        const unsigned rep0_saved = rep0;
-        len = min_match_len + Rd_decode_len( rdec, &d->match_len_model, pos_state );
-        dis_slot = Rd_decode_tree6( rdec, d->bm_dis_slot[get_len_state(len)] );
-        if( dis_slot < start_dis_model ) rep0 = dis_slot;
-        else
+        unsigned distance;
+        len = min_match_len + Rd_decode_len( rdec, &match_len_model, pos_state );
+        distance = Rd_decode_tree6( rdec, bm_dis_slot[get_len_state(len)] );
+        if( distance >= start_dis_model )
           {
+          const unsigned dis_slot = distance;
           const int direct_bits = ( dis_slot >> 1 ) - 1;
-          rep0 = ( 2 | ( dis_slot & 1 ) ) << direct_bits;
+          distance = ( 2 | ( dis_slot & 1 ) ) << direct_bits;
           if( dis_slot < end_dis_model )
-            rep0 += Rd_decode_tree_reversed( rdec,
-                    d->bm_dis + rep0 - dis_slot - 1, direct_bits );
+            distance += Rd_decode_tree_reversed( rdec,
+                        bm_dis + ( distance - dis_slot ), direct_bits );
           else
             {
-            rep0 += Rd_decode( rdec, direct_bits - dis_align_bits ) << dis_align_bits;
-            rep0 += Rd_decode_tree_reversed4( rdec, d->bm_align );
-            if( rep0 == 0xFFFFFFFFU )		/* marker found */
+            distance +=
+              Rd_decode( rdec, direct_bits - dis_align_bits ) << dis_align_bits;
+            distance += Rd_decode_tree_reversed4( rdec, bm_align );
+            if( distance == 0xFFFFFFFFU )		/* marker found */
               {
-              rep0 = rep0_saved;
               Rd_normalize( rdec );
               LZd_flush_data( d );
               if( len == min_match_len )	/* End Of Stream marker */
@@ -298,10 +320,10 @@ int LZd_decode_member( struct LZ_decoder * const d,
               }
             }
           }
-        rep3 = rep2; rep2 = rep1; rep1 = rep0_saved;
+        rep3 = rep2; rep2 = rep1; rep1 = rep0; rep0 = distance;
         state = St_set_match( state );
         if( rep0 >= d->dictionary_size ||
-            ( rep0 >= LZd_data_position( d ) && !d->pos_wrapped ) )
+            ( rep0 >= LZd_data_position( d ) && !d->pos_wrapped_dic ) )
           { LZd_flush_data( d ); return 1; }
         }
       copy_block( d, rep0, len );
